#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include "queensBackend.h"

// Functions for the use case 01. //
static int generateMinQueensNumberConditions
                      (FILE *filePointer, int rows, int columns) {
  int currentRow, currentColumn, valueToWrite;
  int clauses = 0;

  for (currentRow=0; currentRow<rows; currentRow++) {
    for (currentColumn=0; currentColumn<columns; currentColumn++) {
      valueToWrite = currentRow*columns + currentColumn + 1;
      fprintf(filePointer, "%d ", valueToWrite);
    }
    clauses++;
    fprintf(filePointer, "%d\n", 0);
  }
  return clauses;
}

static int generateRowsConditions (FILE *filePointer, int currentRow,
                                    int currentColumn, int rows, int columns) {
  // The current cell is the precedence on the propositional clause:
  // for example, if 1 -> -4 & -7 then currentCell = 1.
  int currentCell, exploredRow, valueToWrite;
  int clauses = 0;

  currentCell = currentRow*columns + currentColumn + 1;
  for (exploredRow=0; exploredRow<rows; exploredRow++) {
    valueToWrite = exploredRow*columns + currentColumn + 1;
    if (currentCell!=valueToWrite) {
      clauses++;
      fprintf(filePointer, "%d %d 0\n", -currentCell, -valueToWrite);
    }
  }
  return clauses;
}

static int generateColumnsConditions (FILE *filePointer, int currentRow,
                                          int currentColumn, int columns) {
  // The current cell is the precedence on the propositional clause:
  // for example, if 1 -> -2 & -3 then currentCell = 1.
  int currentCell, exploredColumn, valueToWrite;
  int clauses = 0;

  currentCell = currentRow*columns + currentColumn + 1;
  for (exploredColumn=0; exploredColumn<columns; exploredColumn++) {
    valueToWrite = currentRow*columns + exploredColumn + 1;
    if (currentCell!=valueToWrite) {
      clauses++;
      fprintf(filePointer, "%d %d 0\n", -currentCell, -valueToWrite);
    }
  }
  return clauses;
}

static int generateDiagonalsConditions (FILE *filePointer, int currentRow,
                                    int currentColumn, int rows, int columns) {
  int currentCell, exploredRow, exploredColumn, valueToWrite;
  int clauses = 0;

  currentCell = currentRow*columns + currentColumn + 1;
  for (exploredRow=0; exploredRow<rows; exploredRow++) {
    for (exploredColumn=0; exploredColumn<columns; exploredColumn++) {
      valueToWrite = exploredRow*columns + exploredColumn + 1;
      if ((currentCell!=valueToWrite)
          && abs(exploredRow-currentRow)==abs(exploredColumn-currentColumn)) {
        clauses++;
        fprintf(filePointer, "%d %d 0\n", -currentCell, -valueToWrite);
      }
    }
  }
  return clauses;
}

static int generateConditions (FILE *filePointer, int rows, int columns) {
  int currentRow, currentColumn;
  int clauses = 0;

  clauses+=generateMinQueensNumberConditions(filePointer, rows, columns);
  for (currentRow=0; currentRow<rows; currentRow++) {
    for (currentColumn=0; currentColumn<columns; currentColumn++) {
      clauses+=generateRowsConditions(filePointer, currentRow,
                                            currentColumn, rows, columns);
      clauses+=generateColumnsConditions(filePointer, currentRow,
                                                  currentColumn, columns);
      clauses+=generateDiagonalsConditions(filePointer, currentRow,
                                                currentColumn, rows, columns);
    }
  }
  return clauses;
}

static void clausesFileHeader (char *path, int variables, int prepositions) {
  char *comment = "c c_i=i where c_i is the cell number i on the matrix.";
  char *first_program_line = "p cnf";
  FILE *filePointer;

  filePointer = fopen(path, "w");
  fprintf(filePointer, "%s\n", comment);
  fprintf(filePointer, "%s %d %d\n", first_program_line,
                                      variables, prepositions);
  fclose(filePointer);
}

static void createSATInputFile (char *sourceFile, char *destFile) {
  char letter;
  int sourcefd, destfd;

  sourcefd = open(sourceFile, O_RDONLY);
  destfd = open(destFile, O_WRONLY);
  lseek(destfd, 0, SEEK_END);
  while (read(sourcefd, &letter, sizeof(char))>0) {
      write(destfd, &letter, sizeof(char));
  }
  remove(sourceFile);
  close(sourcefd);
  close(destfd);
}

// Use cases implemented on the program. //

/* Use case 01 -> write to a file the clauses that explains conditions on
* n-queens problem, using a SAT solver. */
void writeSATsolutionToFile (int rows, int columns) {
  FILE *filePointer;
  int clauses = 0;
  char *sourceFile = "./queens.txt";
  char *destFile = "./queens-clauses.txt";

  filePointer = fopen(sourceFile, "w");
  clauses+=generateConditions(filePointer, rows, columns);
  fclose(filePointer);
  clausesFileHeader(destFile, rows*columns, clauses);
  createSATInputFile(sourceFile, destFile);
}

/* Use case 02 -> read from clasp output a solution for the n-queens problem.
So, this is an example of using a SAT solver. */
void readSATsolutionFromSolverOutput (int rows, int columns) {
  int matrix[rows][columns], currentRow, currentColumn;
  char line[100];
  int value, solution = 0;
  char *solutionPath = "queens-solution.txt";
  FILE *filePointer;

  system("clasp queens-clauses.txt > queens-solution.txt");
  filePointer = fopen(solutionPath, "r");

  while (!feof(filePointer)) {
    fscanf(filePointer, "%s", line);
    if (!strcmp(line, "v")) {
      solution++;
      break;
    }
  }

  while (!feof(filePointer)) {
    fscanf(filePointer, "%s", line);
    if (strcmp(line, "v")) {
      value = atoi(line);
      if (value) {
        currentRow = (abs(value)-1)/rows;
        currentColumn = (abs(value)-1)%columns;
        matrix[currentRow][currentColumn] = value;
      } else {
        break;
      }
    }
  }

  fclose(filePointer);

  if (solution) {
    for (currentRow=0; currentRow<rows; currentRow++) {
      for (currentColumn=0; currentColumn<columns; currentColumn++) {
        if (matrix[currentRow][currentColumn]<0) {
          printf("- ");
        } else {
          printf("Q ");
        }
      }
      printf("\n");
    }
  } else {
    printf("It hasn't got solution.\n");
  }

}
